class Tracer:

    # Initializer and Instance Attributes
    def __init__(self,LabeledElement,TracerEnrichment,LabelEnrichment):
        from pdb import set_trace

        self.LabeledElement = LabeledElement # A string indicating the atom that is considered to be labeled
        self.TracerEnrichment = TracerEnrichment
        self.LabelEnrichment = LabelEnrichment
        self.NaturalLabeledAtomMID = None # The MID of the labeled atom if it were unlabeled (needed when tracer and/or label enrichments are specified)

        self.ReadMID()


    def ReadMID(self):
        # This method opens the text file defining the AtomMIDs and imports that information into the attribute NaturalMID as a numpy array

        import os.path
        import numpy as np
        import PolyMID
        from pdb import set_trace

        PolyMID_Path = os.path.abspath(PolyMID.__file__)
        PolyMID_Path = PolyMID_Path.split(sep='/')
        PolyMID_Path = PolyMID_Path[:-1]

        AtomMIDs_txtPath = '/'.join(PolyMID_Path) + '/SupportingFiles/AtomIMDs.txt'

        with open(AtomMIDs_txtPath,'r') as AtomMIDsFile:
            for line in AtomMIDsFile:
                line_split = line.split(':')
                FileAtomSymbol = line_split[0].strip()
                FileAtomMID_String = line_split[1].strip()

                if FileAtomSymbol == self.LabeledElement:
                    AtomMID_String = FileAtomMID_String
                    AtomMID_StringArray = AtomMID_String.split(sep=' ')
                    AtomMID_FloatArray = [float(i) for i in AtomMID_StringArray]
                    self.NaturalLabeledAtomMID = np.asarray(AtomMID_FloatArray)



def TextToCM(text,CM):
#    This function reads information on CM from a string of text
#    The notation of the text is provided in the example file and is as follows:
#        O[7 2.4 8; 9 4 2; 2 0 6], N[7 2.4 8; 9 4 2; 2 0 6], C[7 2.4 8; 9 4 2; 2 0 6]
#        Will be read as a dictionary with keys for 'O', 'N', and 'C'
#            The correction matrices will be numpy arrays with multiple rows separated by the semicolons
#    Inputs:
#        CM: a dictionary with element symbols as keys
#    Outputs:
#        CM: a dictionary with element symbols as keys and numpy ndarrays containing corretion matrices as values 

    import numpy as np
    from pdb import set_trace

    text_split = text.split(',')
    n_entries = len(text_split)

    for entry in text_split:
        entry = entry.strip()
        atom = entry[0]
        matrix = entry[2:-1]
        rows = matrix.split(';')
        dict_placeholder = dict()
        i = 0
        for row in rows:
            row = row.strip()
            dict_placeholder[i] = np.fromstring(row,sep=' ')
            i = i+1

        #Turn the dictionary into a numpy multidimensional ndarray
        n_keys = len(dict_placeholder.keys())
        n_MID_entries = len(dict_placeholder[0])
        CM[atom] = np.zeros((n_keys,n_MID_entries))
        for key in dict_placeholder.keys():
            CM[atom][key,] = dict_placeholder[key]

    return(CM)



def Correct(CorrectInput=None):

    # Inputs:
    #     CorrectInput:
    #         Can be a fragment object containing information of the fragment whose MID is being corrected for natural isotopic abundances
    #         Can be a string representing a path to a text file contining the information to make a fragment object
    #         Can be None, in which case the user is prompted to select a text file containing the above information using a GUI

    #     Defined within the fragment object:
    #         LabeledElement: A character that is the chemical symbol of the atom which is assumed to be labeled in the fragment whose MID is being corrected for natural isotopic abundances
    #         TracerEnrichment: The percent of the molecule assumed to be the tracer that is actually the tracer (e.g. 50% of glucose is [U-13C6]glucose)
    #         LabelEnrichment: The percent of the atom that is assumed to be labeled that is actually labeled (e.g. 99% of the atoms said to be 13C in [U-13C6]glucose are actually 13C)
    #         HighRes: A  list of strings indicating which elements have heavy isotopes whose mass differences are resolved from the mass differences due to heavy isotopes of the tracer element
    #             can also be 'all' or 'none'
    #             i.e. whether M1, M2, M3 are aggregate measurements of heavy isotopes of all atoms (low resolution, indicated by HighRes='none') or just the atom that is labeled and those not indicated in HighRes
    #             Note: Correcting high resolution data is achieved by setting all atom MIDs, for the elements indicated in HighRes, in the atom objects to [1 0 0] except that of the LabeledElement
    #                 This is accomplished in the definition of an Atom object where atom MIDs for high resolution data are taken from a separate file
    #             Note: this is now entered as a list separated by spaces in the input text file


    # Outputs:
    #     fragment: A fragment object where the corrected MID is stored as a numpy array in fragment.MIDc

    import numpy as np
    import pandas as pd
    import pdb
    from PolyMID import Fragment
    from PolyMID import InputClass
    from PolyMID import Tracer
    from pdb import set_trace

    #Initialize the Inputs variable as an Input object
    InputObject = InputClass(CorrectInput)

    print('\nCalculating the corrected MID...')

    #Create a Fragment object
    #    If a fragment object was passed to this function, this will be equivalent
    fragment = InputObject.fragment

    #Create a correction matrix and calculate its inverse
    fragment.create_correction_matrix()

    #Calculate the corrected MID
    fragment.calc_corrected_mid()

    #Print the corrected MID
    print('The corrected MID is as follows:\n')
    MIDc = pd.DataFrame(fragment.MIDc)
    MIDc = MIDc.transpose()
    nColumns = MIDc.shape[1]
    MIDcColumns = np.array([])
    ColumnNumbersList = list(range(nColumns))
    for i in ColumnNumbersList:
        MIDc = MIDc.rename(columns={i:'M'+ str(i)})
    MIDc.index=[fragment.name]
    pd.set_option("display.precision", 3) #set to print 8 decimals by default
    print(MIDc)
    print('\n\n')

    return(fragment)



class Atom:

    # Initializer and Instance Attributes
    def __init__(self,AtomSymbol,AtomStoich,Tracer,HighRes):
        import numpy as np
        from pdb import set_trace

        self.symbol = AtomSymbol
        self.stoich = AtomStoich # The quantity of the atom in the molecule
        self.MID = None # The isotopic abundances of the atom
        self.Tracer = Tracer # a tracer object with information about the tracer
        self.HighRes = HighRes # a numpy array indicating which elements have heavy isotopes whose mass differences are resolved from the mass differences due to heavy isotopes of the tracer element

        # Retrieve the Atom MID from the text file PolyMID/SupportingFiles/AtomMIDs.txt
        self.ReadMID()

    # instance method to assign new values to an Atom object
    def assign(self,attribute,NewValue):
        if attribute == 'symbol':
            self.symbol = NewValue
        if attribute == 'MID':
            self.MID = NewValue
        if attribute == 'name':
            self.name = NewValue

    def ReadMID(self):
        # This method opens the text file defining the AtomMIDs and imports that information into the attribute MID as a numpy array

        import os.path
        import numpy as np
        import PolyMID
        from PolyMID import PadToEqualLength
        from pdb import set_trace

        PolyMID_Path = os.path.abspath(PolyMID.__file__)
        PolyMID_Path = PolyMID_Path.split(sep='/')
        PolyMID_Path = PolyMID_Path[:-1]

        # If the atom is not part of a fragment measured on a high resolution instrument
        #     Consider its heavy isotopes
        if (self.symbol not in self.HighRes) & ('all' not in self.HighRes):
            AtomMIDs_txtPath = '/'.join(PolyMID_Path) + '/SupportingFiles/AtomIMDs.txt'

        # If the atom is the atom which carries a label (i.e. the one whose mass isotopomers are measured)
        #     Consider its heavy isotopes
        if self.symbol == self.Tracer.LabeledElement:
            AtomMIDs_txtPath = '/'.join(PolyMID_Path) + '/SupportingFiles/AtomIMDs.txt'

        # If the atom is part of a fragment measured on a high resolution instrument and it is not the atom which carries a label (i.e. one whose mass isotopomers are not measured)
        #     Do not consider its heavy isotopes
        if ((self.symbol in self.HighRes)|('all' in self.HighRes)) & (self.symbol != self.Tracer.LabeledElement):
            AtomMIDs_txtPath = '/'.join(PolyMID_Path) + '/SupportingFiles/AtomIMDsHighRes.txt'

        with open(AtomMIDs_txtPath,'r') as AtomMIDsFile:
            for line in AtomMIDsFile:
                line_split = line.split(':')
                FileAtomSymbol = line_split[0].strip()
                FileAtomMID_String = line_split[1].strip()

                if FileAtomSymbol == self.symbol:
                    AtomMID_String = FileAtomMID_String
                    AtomMID_StringArray = AtomMID_String.split(sep=' ')
                    AtomMID_FloatArray = [float(i) for i in AtomMID_StringArray]
                    self.MID = np.asarray(AtomMID_FloatArray)
                    # Consider Tracer and Label enrichment by having the heavy atom be components of the heavy atom MID and the natural atom MID
                    if self.symbol == 'Hv':
                        coefficient = self.Tracer.TracerEnrichment*self.Tracer.LabelEnrichment
                        #    The coefficient is the product of these two enrichments because that provides the probability that an atom considered to be labeled will actually be labeled
                        # The arrays must be the same length to make a linear combination of both of them
                        #     Ensure they are the same length by appending 0s to the shorter one
                        array1 = self.MID
                        array2 = self.Tracer.NaturalLabeledAtomMID
                        array1,array2 = PadToEqualLength(array1,array2)
                        self.MID = coefficient*array1 + (1-coefficient)*array2



def get_directory(retrieve_directory_method):
    from tkinter import Tk #allows for asking for a directory through a GUI
    from tkinter.filedialog import askdirectory #allows for asking for a directory through a GUI
    from tkinter.filedialog import askopenfilename #allows for asking for a file through a GUI

    if retrieve_directory_method == 'gui':
        #ask for the directory where the netCDF and library.txt files are
        root = Tk()
        root.withdraw() #closes the tkinter GUI window because the rest of the program is not run through the GUI
        file_directory = askdirectory() + '/'
        root.update() #required so the directory request dialog box disappears and does not freeze

    if retrieve_directory_method == 'manual':
        #manually enter file directory because tkinter is giving troubles when used with pdb.set_trace()
        file_directory = '/Users/nate/Dropbox/Research/Lehtio_Laboratory/Projects/metabolite_integration_tool/netcdf_test/'

    if retrieve_directory_method == 'gui_file':
        #ask for the directory where the netCDF and library.txt files are
        root = Tk()
        root.withdraw() #closes the tkinter GUI window because the rest of the program is not run through the GUI
        file_directory = askopenfilename()
        root.update() #required so the directory request dialog box disappears and does not freeze

    if retrieve_directory_method == 'manual_file':
        #manually enter file directory because tkinter is giving troubles when used with pdb.set_trace()
        file_directory = '/Users/Nate/Desktop/test_cdfs/TBDMS01_A549NTKD_DMSO.CDF'

    return(file_directory)



def quantity_of_atom(mol_formula,atomic_id):
    import numpy as np
    import re

    #break the metabolite atoms (a formula containing on the atoms that are part of the original metabolite) up so atoms and quantities are consecutive entries in a numpy array
    broken_mol_formula = np.array(re.findall('[A-Z][a-z]?|[0-9]+', mol_formula))
    #    '[A-Z][a-z]?|[0-9]+': A capital letter [A-Z], followed by a lowercase letter [a-z], which is optional '?', or '|' a number '[0-9]', and possibly more numbers '+'
    #    example: this command will take formula = C6H12N1O3Si1 and return broken_formula = array(['C','6','H','12','N','1','O','3','Si','1'])
    #        all components are strings

    #the number of rows of the correction matrix is equal to the quantity of the atom being corrected for that are in the fragment and the original metabolite
    atom_index = np.where(broken_mol_formula==atomic_id)[0][0]
    atom_quantity_index = atom_index+1
    atom_quantity = broken_mol_formula[atom_quantity_index]
    atom_quantity = atom_quantity.astype(np.int)

    return(atom_quantity)



def expand_polynomial(a,b):

# Inputs:
#     a: a python array representing the mass isotopomer abundances of atom a
#        the first entry is the abundance of M0, the second M1, the third M2, ...
#     b: a python array representing the mass isotopomer abundances of atom b
#        the first entry is the abundance of M0, the second M1, the third M2, ...
# Outputs:
#     a Pandas data frame is returned where the rows correspond to mass isotopomers and the column contains the mass isotopomer abundances as 'prob'
#         This retunred mass isotopomer distribution (MID) is for the molecule comprised of the atoms ab
#         Note this can be used in series to get a multi-atom molecule MID
#             Say you have the molecule abc
#             First run this function with the MIDs for the atoms a and b as inputs; the output is then the MID for the combined ab
#             Run the function again with the inputs being the previous output and the MID for c; the output is then the MID for abc

    import numpy as np
    import pandas
    import importlib #allows fresh importing of modules
    from pdb import set_trace

    # Redefine the input polynomial python arrays as numpy row matrices
    mid_a = np.matrix(a) #mass isotopomer distribution for a
    mid_b = np.matrix(b) #mass isotopomer distribution for b

    #find the number of recorded relative mass isotopomer abundances for each atom (the number of columns)
    n_a = mid_a.shape[1] #the second entry of the returned tuple
    n_b = mid_b.shape[1] #the second entry of the returned tuple


    # Create Pandas data frames to hold the mass isotopoer relative abundance and the number of extra neutrons associated with that mass isotopomer in separate columns
    #     The columns are 'prob' and 'm_isotopomer' respectively

    # Arrange information for atom a into a pandas data frame
    mid_a_m = np.zeros([n_a,2]) #initialize the matrix the size of data frame that will be used to hold the information for atom a
    mid_a_df = pandas.DataFrame(mid_a_m,columns=['prob','m_isotopomer']) #initialize the data frame containing the information for atom a
    mid_a_df[['prob']] = np.transpose(mid_a) #enter the relative abundances of each mass isotopomer
    mid_a_df[['m_isotopomer']] = np.transpose(np.matrix(range(0,n_a))) #enter the mz mass isotopomer values corresponding to each mass isotopomer

    # Arrange information for atom b into a pandas DataFrame
    mid_b_m = np.zeros([n_b,2])
    mid_b_df = pandas.DataFrame(mid_b_m,columns=['prob','m_isotopomer'])
    mid_b_df[['prob']] = np.transpose(mid_b)
    mid_b_df[['m_isotopomer']] = np.transpose(np.matrix(range(0,n_b)))

    # Initialize a DataFrame to contain information on all possible combinations of atoms a and b in the molecule ab
    factored_init_m = np.zeros([n_a*n_b,2])
    factored = pandas.DataFrame(factored_init_m,columns=['prob','m_isotopomer'])

    # Fill in the factored_df by expanding the polynomial and keeping track of the mass isotopomer mz value (M0,M1,M2,...) for each term
    #     This is expanding the polynomial and storing each term
    #     The mass isotopomer mz values of each factor are summed for each term to get a new mass isotopomer mz value for each term

    # Initialize the variable to track the row of the factored data frame
    #     This iterates through every possible combination of entries in mid_a_df and mid_b_df
    iterator = 0
    for i in range(0,n_a):
        for k in range(0,n_b):
            # calculate the probalility of a combiniation of mass isotopomers from atom a and b
            factored.loc[iterator,'prob'] = mid_a_df.loc[i,'prob']*mid_b_df.loc[k,'prob']
            # determine the value of the mass isotomer of that combination (i.e. M0, M1, or M2, etc.)
            factored.loc[iterator,'m_isotopomer'] = mid_a_df.loc[i,'m_isotopomer'] + mid_b_df.loc[k,'m_isotopomer']
            iterator = iterator + 1

    # Combine the relative mass isotopmer abundances for ab combinations with the same mass isotopomer mz value
    grouped = factored.groupby('m_isotopomer')
    #     the groupby method of a Pandas data frame returns a group object
    #         a grouped object has the aggregate method that takes a function as an input and returns a Pandas data frame
    factored_aggregated = grouped.aggregate(np.sum)

    # Shorten factored to remove negligible values
    #     Removes all values starting from the end of the MID that are less than the specified cut-off
    cut_off = 1e-4
    factored_aggregated_array = np.array(factored_aggregated)
    factored_aggregated_array[factored_aggregated_array < cut_off] = 0
    factored_aggregated_trimmed = np.trim_zeros(factored_aggregated_array,'b')
    n_factors = len(factored_aggregated_trimmed)
    factored_aggregated = factored_aggregated.iloc[0:n_factors]

    # Return the Pandas data frame containing the mass isotopomer distrubution of the molecule ab
    return(factored_aggregated)



def PadToEqualLength(array1,array2):
#     Inputs:
#     array1: an array whose length is to be made equal to that of array2 by lengthing the shorter of the two arrays by padding with 0s at the end
#     array2: an array whose length is to be made equal to that of array1 by lengthing the shorter of the two arrays by padding with 0s at the end

    # Import required packages
    import numpy as np

    len1 = len(array1)
    len2 = len(array2)

    n_zeros_needed = abs(len2 - len1)

    if len1 < len2:
        array1 = np.pad(array1,(0,n_zeros_needed),mode='constant')

    if len2 < len1:
        array2 = np.pad(array2,(0,n_zeros_needed),mode='constant')

    return(array1,array2)



class Fragment:

    # Initializer and Instance Attributes
    def __init__(self,FragmentName,FragmentFormula,CanAcquireLabel,MIDm,LabeledElement,TracerEnrichment,LabelEnrichment,HighRes,MIDc=None,PeakArea=None,CM=None):

        from PolyMID import Formula
        from PolyMID import Tracer
        import numpy as np
        from pdb import set_trace

        # MIDm and HighRes may be passed as python lists
        #     They must be converted to numpy arrays
        MIDm = np.array(MIDm)
        HighRes = np.array(HighRes)

        self.name = FragmentName
        self.MIDm = MIDm
        self.MIDc = MIDc
        self.CM = CM
        self.CMi = None
        self.PeakArea = PeakArea
        self.Tracer = Tracer(LabeledElement,TracerEnrichment,LabelEnrichment)
        self.HighRes = HighRes # A numpy array indicating which elements have heavy isotopes whose mass differences are resolved from the mass differences due to heavy isotopes of the tracer element
        self.Formula = Formula(formula=FragmentFormula,Tracer=self.Tracer,HighRes=self.HighRes)
        self.CanAcquireLabel = Formula(formula=CanAcquireLabel,Tracer=self.Tracer,HighRes=self.HighRes)

    # instance method to assign new values to a Fragment object
    def assign(self,attribute,NewValue):
        if attribute == 'name':
            self.name = NewValue
        if attribute == 'formula':
            self.formula = NewValue
        if attribute == 'CanAcquireLabel':
            self.CanAcquireLabel = NewValue
        if attribute == 'MIDm':
            self.MIDm = NewValue
        if attribute == 'MIDc':
            self.MIDc = NewValue
        if attribute == 'CM':
            self.CM = NewValue
        if attribute == 'CMi':
            self.CMi = NewValue
        if attribute == 'PeakArea':
            self.PeakArea = NewValue

    # instance method to create correction matrix for a Fragment object
    def create_correction_matrix(self):

        from pdb import set_trace
        import numpy as np
        import re
        import copy
        from PolyMID import quantity_of_atom
        from PolyMID import Formula

        #break the formula up so atoms and quantities are consecutive entries in a numpy array
        broken_formula = np.array(re.findall('[A-Z][a-z]?|[0-9]+', self.Formula.formula))
        #    '[A-Z][a-z]?|[0-9]+': A capital letter [A-Z], followed by a lowercase letter [a-z], which is optional '?', or '|' a number '[0-9]', and possibly more numbers '+'
        #    example: this command will take formula = C6H12N1O3Si1 and return broken_formula = array(['C','6','H','12','N','1','O','3','Si','1'])
        #        all components are strings
        n_formula_entries = len(broken_formula)

        #find the index of the number of the atom which can acquire a label in the formula for the full fragment
        #    it is used in creating the correction matrix below because successive quantities of this atom need to be subtracted and a heavy atom put in its place
        atom_index = np.where(broken_formula==self.Tracer.LabeledElement)[0][0]
        atom_quantity_index = atom_index+1 #refering to full fragment

        #the number of rows of the correction matrix is equal to the quantity of the atom being corrected for that are in the fragment and the original metabolite
        atom_quantity = quantity_of_atom(self.CanAcquireLabel.formula,self.Tracer.LabeledElement) #this does not refer to the full fragment!

        #add the "heavy atom to the end of the broken formula array", initially its quantity is 0
        broken_formula = np.append(broken_formula,np.array(['Hv','0']))
        n_formula_entries_heavy = len(broken_formula)

        #replace each atom of interest with a heavy atom and get the natural mid of the result
        #    these mids fill the rows of the correction matrix
        broken_formula_correct = copy.copy(broken_formula) #initialize the array to carry the formula with a heavy atom
        correction_matrix_dict = dict() #initialize a dictionary to hold the rows of the correction matrix
        for i in range(0,atom_quantity+1):
            #subtract an atom of interest from the formula
            broken_formula_correct[atom_quantity_index] = broken_formula[atom_quantity_index].astype(np.int) - i
            broken_formula_correct[atom_quantity_index] = broken_formula_correct[atom_quantity_index].astype(np.str)

            #replace that atom with a heavy atom
            broken_formula_correct[n_formula_entries+1] = broken_formula[n_formula_entries+1].astype(np.int) + i
            broken_formula_correct[n_formula_entries+1] = broken_formula_correct[n_formula_entries+1].astype(np.str)

            #update the string version of the formula from the array version
            new_formula = ''
            for j in range(0,n_formula_entries_heavy):
                new_formula = new_formula + broken_formula_correct[j]

            #make a Formula object
            new_formula = Formula(formula=new_formula,Tracer=self.Tracer,HighRes=self.HighRes)

            #get the mid due to natural abundances of the updated formula (with one or more heavy atoms)
            new_formula.calc_natural_mid()
            correction_matrix_dict[i] = new_formula.NaturalMID

            # Keep track of the length of the longest natural MID with heavy atoms
            #     All of the MIDs of the molecule with incorporated heavy atoms must be set to the same length to form a correction matrix
            if i==0:
                n_theoretical_mid_entries = len(correction_matrix_dict[i])

            if i>0:
                if n_theoretical_mid_entries < len(correction_matrix_dict[i]):
                    n_theoretical_mid_entries = len(correction_matrix_dict[i])

        # Make all of the natural MIDs of the molecule with incorporated heavy atoms the same length by appending zeros
        #     This length is referred to as the number of theoretical MID entries
        for i in range(0,atom_quantity+1):
            # The number of theoretical MID entries in the correction matrix must be the same as the number of measured MID entries for the matrix algebra to work
            #     If the number of theoretical MID entries is less, it is increased
            #     The case where n_theoretical_mid_entries is greater (the usual case) is handled later by increasing the number of measured MID entries
            if n_theoretical_mid_entries < len(self.MIDm):
                n_theoretical_mid_entries = len(self.MIDm)

            # Lengthen each theoretical MID with given quantities of heavy atoms to the specified length of the theoretical MIDs
            #     MIDs will NOT have to be shortened here because they are all set to have the same length as the longest theoretical MID
            if len(correction_matrix_dict[i]) < n_theoretical_mid_entries:
                n_zeros_needed = n_theoretical_mid_entries - len(correction_matrix_dict[i])
                correction_matrix_dict[i] = np.pad(correction_matrix_dict[i],(0,n_zeros_needed),mode='constant')

        #make the correction matrix dictionary into a matrix
        n_keys = len(correction_matrix_dict.keys())
        n_MID_entries = len(correction_matrix_dict[0])
        CM = np.zeros((n_MID_entries,n_keys))
        CM_rows = np.arange(0,n_MID_entries)
        for key in correction_matrix_dict.keys():
            CM[CM_rows,key] = correction_matrix_dict[key]

        #find the inverse of the correction matrix
        #    Use the Moore–Penrose pseudo-inverse because the matrix is not necessarily square and the linear problem is "over specified"
        CMi = np.linalg.pinv(CM)
        #    CMi = np.linalg.inv(CM) would give the inverse of a square matrix


        self.CM = CM
        self.CMi = CMi

    def calc_corrected_mid(self):

        import numpy as np
        from pdb import set_trace

        #find the inverse of the correction matrix
        CMi = self.CMi

        #the theoretical MIDs are the rows of the correction matrix.
        #    Their length corresponds to the number of rows of the right inverse of the correcion matrix
        n_theoretical_mid_entries = CMi.shape[1]

        #the measured MID must have the same number of entries as each of the theoretical MIDs
        #    if it is short, add zeros to make up for the difference
        MIDm = self.MIDm
        if len(MIDm) < n_theoretical_mid_entries:
            mid_u_appendage = np.zeros(n_theoretical_mid_entries-len(MIDm))
            MIDm = np.append(MIDm,mid_u_appendage)

        #calculate corrected MID
        MIDc = np.dot(CMi,MIDm)
        MIDc = MIDc/sum(MIDc)
        self.assign('MIDc',MIDc)



class InputClass:
    #Initializer a Formula Instance and its attributes
    def __init__(self,CorrectInput):
        from pdb import set_trace
        from PolyMID import Fragment
        from PolyMID import Tracer
        import numpy as np
        from PolyMID import get_directory
        from PolyMID import TextToCM

        # If the input to the Correct() function is a fragment object, than it can be used directly.
        # Otherwise a fragment object must be populated by reading from a text file.
        ReadFromTextFile = (isinstance(CorrectInput,str)) | (CorrectInput is None)

        if not ReadFromTextFile:
            self.fragment=CorrectInput

        if ReadFromTextFile:

            #If there is no input provided, the the user is prompted to select a text file using a GUI.
            if (CorrectInput is None):
                CorrectInput = get_directory('gui_file')

            #Initialize variables
            FragmentName = None
            formula = None
            CanAcquireLabel = None
            MIDm = None
            MIDc = None
            CM = None

            #import values from text file
            with open(CorrectInput, 'r') as read_file:
                for line in read_file:
                    line_split = line.split(':')
                    line_split[0] = line_split[0].strip()
                    line_split[1] = line_split[1].strip()

                    if (line_split[0] == 'FragmentFormula') | (line_split[0] == 'Fragment Formula'):
                        FragmentFormula = line_split[1]

                    if (line_split[0] == 'CanAcquireLabel') | (line_split[0] == 'Metabolite Atoms'):
                        CanAcquireLabel = line_split[1]

                    if (line_split[0] == 'MIDm'):
                        MIDm = line_split[1]
                        MIDm = np.fromstring(MIDm,dtype=float,sep=' ')

                    if (line_split[0] == 'FragmentName') | (line_split[0] == 'Fragment Name'):
                        FragmentName = line_split[1]

                    if (line_split[0] == 'LabeledElement') | (line_split[0] == 'Labeled Element'):
                        LabeledElement = line_split[1]

                    if (line_split[0] == 'TracerEnrichment') | (line_split[0] == 'Tracer Enrichment'):
                        TracerEnrichment = float(line_split[1])

                    if (line_split[0] == 'LabelEnrichment') | (line_split[0] == 'Label Enrichment'):
                        LabelEnrichment = float(line_split[1])

                    if (line_split[0] == 'HighRes') | (line_split[0] == 'High Res'):
                        HighRes = line_split[1]
                        if HighRes == 'none':
                            HighRes = np.array([],dtype='str')
                        # convert to a list of the elements that are resolved with high resolution
                        if (HighRes!='none') & (HighRes!='all'):
                            HighRes = HighRes.strip().split(' ')

            self.fragment = Fragment(FragmentName=FragmentName, FragmentFormula=FragmentFormula, CanAcquireLabel=CanAcquireLabel, MIDm=MIDm, LabeledElement=LabeledElement, TracerEnrichment=TracerEnrichment, LabelEnrichment=LabelEnrichment, HighRes=HighRes, MIDc=None, PeakArea=None, CM=CM)



def Natural(formula,LabeledElement='C',HighRes=False):
    # Inputs:
    # formula: a string representing the formula whose natural MID is being calculated
    # LabeledElement: A character that is the chemical symbol of the atom which is assumed to be labeled in the fragment whose MID is being corrected for natural isotopic abundances
    # HighRes: A booleon indicating whether the data is high resoloution or not
    #     i.e. whether M1, M2, M3 are aggregate measurements of heavy isotopes of all atoms (not HighRes) or just the atom that is labeled (HighRes)
    #     This input is only used if the fragment information is read from a text file
    #         Otherwise HighRes is already an attribute of the fragment input
    #     Note: Correcting high resolution data is achieved by setting all atom MIDs in the atom objects to [1 0 0] except that of the LabeledElement
    #         This is accomplished in the definition of an Atom object where atom MIDs for high resolution data are taken from a separate file

    # Outputs:
    # formula.NaturalMID: the MID corresponding to the chemical formula resulting from natural heavy isotope abundances

    #Import necessary packages and object classes
    from PolyMID import Formula
    from pdb import set_trace

    #Create a formula object
    formula = Formula(formula=formula,LabeledElement=LabeledElement,HighRes=HighRes)
    #     The Formula object designation takes LabeledElement and HighRes inputs
    #         If HighRes is False, LabeledElement does not matter (which is default behaviour)
    #         High resolution natural MIDs can be calculated by specifying HighRes and LabeledElement, where LabeledElement is the atom contributing heavy isotopes

    #Determine the NaturalMID attribute of the formula object using the calc_natural_mid() method
    formula.calc_natural_mid()

    #Return the natural MID
    return(formula.NaturalMID)



class Formula:
    # Initializer a Formula Instance and its attributes
    def __init__(self,formula,Tracer,HighRes):
        self.FormulaInput = formula
        self.formula = None
        self.AtomArray = None
        self.NaturalMID = None
        self.Tracer = Tracer # A Tracer object with information about the tracer
        self.HighRes = HighRes # A numpy array indicating which elements have heavy isotopes whose mass differences are resolved from the mass differences due to heavy isotopes of the tracer element
        self.FormatFormula()
        self.CreateAtomArray()

    def FormatFormula(self):
        # This method redefines the formula attribute of the Formula class
        #     This method adds 1s (ones) where they are implied to the input string representing the chemical formula (self.FormulaInput)

        import re
        from pdb import set_trace

        # Find the flanking characters where a 1 should be inserted (1s that are internal to the formula)
        FormulaToEdit = self.FormulaInput
        missing1s = re.findall('(?:[A-Z]|[a-z])[A-Z]',FormulaToEdit)
        #    '(?:)' in '(?:RegExpHere)' denotes a non-capturing group, i.e. it is used for specifying order of operations as opposed to just '()'

        # Insert the internal 1s
        for entry in missing1s:
            NewEntry = entry[0] + '1' + entry[1]
            FormulaToEdit = re.sub(entry,NewEntry,FormulaToEdit)

        # Append an external 1 if necessary
        #     If there is a letter at the end of the string, replace it with that letter and a 1
        LetterAtEnd = re.findall('(?:[A-Z]|[a-z])$',FormulaToEdit)
        #    '(?:)' in '(?:RegExpHere)' denotes a non-capturing group, i.e. it is used for specifying order of operations as opposed to just '()'
        if (len(LetterAtEnd) > 0):
            FormulaToEdit = re.sub(LetterAtEnd[0]+'$',LetterAtEnd[0]+'1',FormulaToEdit)

        # Redefine the formula attribute of the Formula obeject
        self.formula = FormulaToEdit

    def CreateAtomArray(self):
        # This method creates an array of atom objects corresponding to the formula

        # Import required modules
        from pdb import set_trace
        import numpy as np
        import re
        from PolyMID import Atom

        # Break the fragment formula up into its atomic symbol and letter components
        broken_formula = np.array(re.findall('[A-Z][a-z]?|[0-9]+', self.formula))
        #    '[A-Z][a-z]?|[0-9]+': A capital letter [A-Z], followed by a lowercase letter [a-z], which is optional '?', or '|' a number '[0-9]', and possibly more numbers '+'
        #    example: this command will take formula = C6H12N1O3Si1 and return broken_formula = array(['C','6','H','12','N','1','O','3','Si','1'])
        #        all components are strings

        # Get separate arrays containing the atomic symbols and corresponding formula numbers
        odd_indices = np.array(range(1,len(broken_formula),2)) #all of the odd indices of the array broken_formula
        even_indices = np.array(range(0,len(broken_formula),2)) #all of the even indices of the array broken_formula
        formula_numbers = broken_formula[odd_indices].astype(np.int) #the atomic numbers array
        formula_atoms = broken_formula[even_indices] #the atomic symbols array

        AtomArray = []
        AtomCounter = 0
        for atom in formula_atoms:
            AtomStoich = formula_numbers[AtomCounter]
            AtomArray.append(Atom(atom,AtomStoich,Tracer=self.Tracer,HighRes=self.HighRes))
            AtomCounter = AtomCounter + 1

        self.AtomArray = AtomArray


    # Method for calculating natural MID of a Formula object
    def calc_natural_mid(self):
        # This method defines the NatrualMID attribute of the Formula class
        # self.formula: the formula for the fragment whose natural isotopic abundance is being calculated
        #    each atomic symbol must start with a capital letter, if there is a second letter it must be lowercase
        #    if there is no number following an atomic symbol, that number is assumed to be 1

        # Import required modules
        from pdb import set_trace
        import numpy as np
        from PolyMID import expand_polynomial

        # Perform the polynomial expansion; expanding and collecting like terms for two MIDs (atoms) at a time
        expanded_placeholder = np.array([1])
        for AtomObject in self.AtomArray:
            for stoich in list(range(AtomObject.stoich)):
                expanded = expand_polynomial(expanded_placeholder,AtomObject.MID)
                expanded_placeholder = np.array(expanded['prob'])

        # Define the NaturalMID attribute of the Formula object
        natural_mid = np.array(expanded['prob'])
        self.NaturalMID = natural_mid



# __init__.py
from .SupportingFunctions.get_directory import get_directory
from .SupportingFunctions.expand_polynomial import expand_polynomial
from .SupportingFunctions.quantity_of_atom import quantity_of_atom
from .SupportingFunctions.TextToCM import TextToCM
from .SupportingFunctions.PadToEqualLength import PadToEqualLength
from .Classes.FragmentClass import Fragment
from .Classes.FormulaClass import Formula
from .Classes.InputClass import InputClass
from .Classes.AtomClass import Atom
from .Classes.TracerClass import Tracer
from PolyMID.Correct import Correct
from PolyMID.Natural import Natural
